/*  ----------------------------------------------------------------<Prolog>-
    Name:       sfluid.c
    Title:      Process user id (uid) and group id (gid) functions
    Package:    Standard Function Library (SFL)

    Written:    96/05/03  iMatix SFL project team <sfl@imatix.com>
    Revised:    98/06/02

    Copyright:  Copyright (c) 1991-98 iMatix
    License:    This is free software; you can redistribute it and/or modify
                it under the terms of the SFL License Agreement as provided
                in the file LICENSE.TXT.  This software is distributed in
                the hope that it will be useful, but without any warranty.
 ------------------------------------------------------------------</Prolog>-*/

#include "prelude.h"                    /*  Universal header file            */
#include "sflstr.h"                     /*  String functions                 */
#include "sfluid.h"                     /*  Prototypes for functions         */


/*  Local constants and function prototypes                                  */

#define UID_CACHE_MAX       10          /*  Max. different uid's we cache    */
#define GID_CACHE_MAX       10          /*  Max. different gid's we cache    */
#define REAL_ID             0           /*  Arguments for get_uid/get_gid    */
#define EFFECTIVE_ID        1

#if (defined (DOES_UID))                /*  Only if uid/gid implemented      */
#   if (!defined (__OS2__))             /*  But not needed under OS/2        */
static uid_t  get_uid (int type);
static gid_t  get_gid (int type);
#   endif
#endif


/*  ---------------------------------------------------------------------[<]-
    Function: get_uid_name

    Synopsis:
    Get user name from passwd file.  We optimise by keeping a table of uids
    and names in memory.  Note that this will cause problems if the program
    stays running when the passwd file has been changed.  Returns a string
    containing the translated user name, or "<none>" if the uid could not
    be translated.  Under MS-DOS the uid must be zero.  The returned string
    is in a static area that is _not_ overwritten with each call, but which
    should be treated as read-only, and unstable: i.e. the value returned
    by one call to get_uid_name may change as a result of a later call.  If
    you need persistent strings, use strdupl() after each call.
    ---------------------------------------------------------------------[>]-*/

char *
get_uid_name (uid_t uid)
{
#   if (defined (DOES_UID))
    static struct uids {                /*  Table of cached uids             */
        uid_t id;
        char  *name;
    } cache [UID_CACHE_MAX];
    static int
        cache_size = 0,                 /*  Number of uid's in cache         */
        cache_oldest = 0;               /*  Oldest entry in cache            */
    int
        cache_scan;                     /*  Scan through cache               */
    struct passwd
        *passwd_entry;

    /*  First, look for uid in cache                                         */
    for (cache_scan = 0; cache_scan < cache_size; cache_scan++)
        if (cache [cache_scan].id == uid)
            return (cache [cache_scan].name);

    /*  Add new name to cache: if cache was full, kick-out oldest entry      */
    if (cache_size == UID_CACHE_MAX)
      {
        cache_scan = cache_oldest++;
        cache_oldest %= UID_CACHE_MAX;
        free (cache [cache_scan].name);
      }
    else
        cache_scan = cache_size++;

    cache [cache_scan].id = uid;
#   if (defined (__VMS__) && !defined (__VMS_XOPEN))
    cache [cache_scan].name = "<none>";
#   else
    if ((passwd_entry = getpwuid (uid)) == NULL)
        cache [cache_scan].name = "<none>";
    else
        cache [cache_scan].name = strdupl (passwd_entry-> pw_name);
#   endif

    return (cache [cache_scan].name);

#   elif (defined (__MSDOS__))
    return (uid == 0? "user": "<none>");

#   endif
}


/*  ---------------------------------------------------------------------[<]-
    Function: get_gid_name

    Synopsis:
    Get group name from group file.  We optimise by keeping a table of gids
    and names in memory.  Note that this will cause problems if the program
    stays running when the group file has been changed.  Returns a string
    containing the translated user name, or "<none>" if the gid could not
    be translated.  Under MS-DOS the gid must be zero.  The returned string
    is in a static area that is _not_ overwritten with each call, but which
    should be treated as read-only, and unstable: i.e. the value returned
    by one call to get_gid_name may change as a result of a later call.  If
    you need persistent strings, use strdupl() after each call.
    ---------------------------------------------------------------------[>]-*/

char *
get_gid_name (gid_t gid)
{
#   if (defined (DOES_UID))
    static struct gids {                /*  Table of cache'd gids            */
        gid_t id;
        char  *name;
    } cache [GID_CACHE_MAX];
    static int
        cache_size = 0,                 /*  Number of gid's in cache         */
        cache_oldest = 0;               /*  Oldest entry in cache            */
    int
        cache_scan;                     /*  Scan through cache               */
    struct group
        *group_entry;

    /*  First, look for gid in cache                                         */
    for (cache_scan = 0; cache_scan < cache_size; cache_scan++)
        if (cache [cache_scan].id == gid)
            return (cache [cache_scan].name);

    /*  Add new name to cache: if cache was full, kick-out oldest entry      */
    if (cache_size == GID_CACHE_MAX)
      {
        cache_scan = cache_oldest++;
        cache_oldest %= GID_CACHE_MAX;
        free (cache [cache_scan].name);
      }
    else
        cache_scan = cache_size++;

    cache [cache_scan].id = gid;
#   if (defined (__VMS__) && !defined (__VMS_XOPEN))
        cache [cache_scan].name = "<none>";
#   else
    if ((group_entry = getgrgid (gid)) == NULL)
        cache [cache_scan].name = "<none>";
    else
        cache [cache_scan].name = strdupl (group_entry-> gr_name);
#   endif

    return (cache [cache_scan].name);

#   elif (defined (__MSDOS__))
    return (gid == 0? "group": "<none>");

#   endif
}


/*  ---------------------------------------------------------------------[<]-
    Function: set_uid_user

    Synopsis: This function can be used by 'setuid' programs; i.e. programs
    that run under a fixed uid such as 'root'.  Typically such programs need
    to access root resources, but user data files.  To do this they must
    switch between the 'root' uid and the 'user' uid.  This function switches
    to the real user id.  Use set_uid_root() to switch (back) to the 'root'
    uid.  See also: set_gid_user() and set_gid_root().
    ---------------------------------------------------------------------[>]-*/

int
set_uid_user (void)
{
#if (defined (DOES_UID))
#   if (defined (__UTYPE_HPUX))
    return (setuid (get_uid (REAL_ID)));
#   elif (defined (__OS2__))            /*  OS/2 only supports one UID       */
    return (0);
#   elif (defined (__VMS__) && !defined (__VMS_XOPEN))
    return (0);
#   else
    return (seteuid (get_uid (REAL_ID)));
#   endif
#else
    return (0);
#endif
}


/*  ---------------------------------------------------------------------[<]-
    Function: set_uid_root

    Synopsis: This function can be used by 'setuid' programs; i.e. programs
    that run under a fixed uid such as 'root'.  Typically such programs need
    to access root resources, but user data files.  To do this they must
    switch between the 'root' uid and the 'user' uid.  This function switches
    back to the root user id.  Use set_uid_user() to switch to the 'user'
    uid.  See also: set_gid_user() and set_gid_root().
    ---------------------------------------------------------------------[>]-*/

int
set_uid_root (void)
{
#if (defined (DOES_UID))
#   if (defined (__UTYPE_HPUX))
    return (setuid (get_uid (EFFECTIVE_ID)));
#   elif (defined (__OS2__))            /*  OS/2 only supports one UID       */
    return (0);
#   elif (defined (__VMS__) && !defined (__VMS_XOPEN))
    return (0);
#   else
    return (seteuid (get_uid (EFFECTIVE_ID)));
#   endif
#else
    return (0);
#endif
}


#if (defined (DOES_UID) && !defined (__OS2__))
/*  -------------------------------------------------------------------------
    Function: get_uid_id -- internal

    Synopsis: Returns the real (REAL_ID) or effective (EFFECTIVE_ID) uid.
    These values are loaded the first time that the function is called: you
    should not rely on the effective uid after changing the uid.
    -------------------------------------------------------------------------*/

static uid_t
get_uid (int type)
{
    static int
        ruid = 0,
        euid = 0;

    if (ruid == 0)
        ruid = getuid ();
    if (euid == 0)
#   if (defined (__UTYPE_HPUX))
        euid = getuid ();
#   else
        euid = geteuid ();
#   endif

    if (type == REAL_ID)
        return (ruid);
    else
    if (type == EFFECTIVE_ID)
        return (euid);
    else
        return (0);
}
#endif


/*  ---------------------------------------------------------------------[<]-
    Function: set_gid_user

    Synopsis: This function can be used by 'setgid' programs; i.e. programs
    that run under a fixed gid such as 'root'.  Typically such programs need
    to access root resources, but user data files.  To do this they must
    switch between the 'root' gid and the 'user' gid.  This function switches
    to the real user id.  Use set_gid_root() to switch (back) to the 'root'
    gid.  See also: set_uid_user() and set_uid_root().
    ---------------------------------------------------------------------[>]-*/

int
set_gid_user (void)
{
#if (defined (DOES_UID))
#   if (defined (__UTYPE_HPUX))
    return (setgid (get_gid (REAL_ID)));
#   elif (defined (__OS2__))            /*  OS/2 only supports one UID       */
    return (0);
#   elif (defined (__VMS__) && !defined (__VMS_XOPEN))
    return (0);
#   else
    return (setegid (get_gid (REAL_ID)));
#   endif
#else
    return (0);
#endif
}


/*  ---------------------------------------------------------------------[<]-
    Function: set_gid_root

    Synopsis: This function can be used by 'setgid' programs; i.e. programs
    that run under a fixed gid such as 'root'.  Typically such programs need
    to access root resources, but user data files.  To do this they must
    switch between the 'root' gid and the 'user' gid.  This function switches
    back to the root user id.  Use set_gid_user() to switch to the 'user'
    gid.  See also: set_gid_user() and set_gid_root().
    ---------------------------------------------------------------------[>]-*/

int
set_gid_root (void)
{
#if (defined (DOES_UID))
#   if (defined (__UTYPE_HPUX))
    return (setgid (get_gid (EFFECTIVE_ID)));
#   elif (defined (__OS2__))            /*  OS/2 only supports one UID       */
    return (0);
#   elif (defined (__VMS__) && !defined (__VMS_XOPEN))
    return (0);
#   else
    return (setegid (get_gid (EFFECTIVE_ID)));
#   endif
#else
    return (0);
#endif
}


#if (defined (DOES_UID) && !defined (__OS2__))
/*  -------------------------------------------------------------------------
    Function: get_gid_id -- internal

    Synopsis: Returns the real (REAL_ID) or effective (EFFECTIVE_ID) gid.
    These values are loaded the first time that the function is called: you
    should not rely on the effective gid after changing the gid.
    -------------------------------------------------------------------------*/

static gid_t
get_gid (int type)
{
    static int
        rgid = 0,
        egid = 0;

    if (rgid == 0)
        rgid = getgid ();
    if (egid == 0)
#   if (defined (__UTYPE_HPUX))
        egid = getgid ();
#   else
        egid = getegid ();
#   endif

    if (type == REAL_ID)
        return (rgid);
    else
    if (type == EFFECTIVE_ID)
        return (egid);
    else
        return (0);
}
#endif


/*  ---------------------------------------------------------------------[<]-
    Function: get_login

    Synopsis: Returns the identity of the currently-logged user.  The
    returned string is in a static buffer.  Returns NULL if no user is
    currently logged-in.
    ---------------------------------------------------------------------[>]-*/

char *
get_login (void)
{
#if (defined (__WINDOWS__))
#   if (defined (WIN32))
    ULONG
        user_name_max = 255;
    static char
        user_name [256];

    WNetGetUser (NULL, user_name, &user_name_max);
    return (user_name);
#   else
    return ("none");
#   endif
#elif (defined (__UNIX__) || defined (__OS2__))
    return (getlogin ());
#endif
}
